@@ -22,7 +22,7 @@
 #endif
 
 #ifndef QCOROCORE_DEPRECATED
-#  define QCOROCORE_DEPRECATED 
+#  define QCOROCORE_DEPRECATED __attribute__ ((__deprecated__))
 #endif
 
 #ifndef QCOROCORE_DEPRECATED_EXPORT
