@@ -17,6 +17,7 @@
 #include <QEventLoop>
 #include <QObject>
 #include <QCoreApplication>
+#include <QPointer>
 
 namespace QCoro {
 
@@ -41,8 +42,8 @@ public:
      * \param[in] awaitingCoroutine handle of the coroutine that is co_awaiting the current
      * coroutine (continuation).
      */
-    explicit TaskFinalSuspend(std::coroutine_handle<> awaitingCoroutine)
-        : mAwaitingCoroutine(awaitingCoroutine) {}
+    explicit TaskFinalSuspend(const std::vector<std::coroutine_handle<>> &awaitingCoroutines)
+        : mAwaitingCoroutines(awaitingCoroutines) {}
 
     //! Returns whether the just finishing coroutine should do final suspend or not
     /*!
@@ -69,9 +70,10 @@ public:
     void await_suspend(std::coroutine_handle<Promise> finishedCoroutine) noexcept {
         auto &promise = finishedCoroutine.promise();
 
-        if (promise.mResumeAwaiter.exchange(true, std::memory_order_acq_rel)) {
-            promise.mAwaitingCoroutine.resume();
+        for (auto &awaiter : mAwaitingCoroutines) {
+            awaiter.resume();
         }
+        mAwaitingCoroutines.clear();
 
         // The handle will be destroyed here only if the associated Task has already been destroyed
         if (promise.setDestroyHandle()) {
@@ -87,8 +89,7 @@ public:
     constexpr void await_resume() const noexcept {}
 
 private:
-    //! Handle of the coroutine co_awaiting the current coroutine.
-    std::coroutine_handle<> mAwaitingCoroutine;
+    std::vector<std::coroutine_handle<>> mAwaitingCoroutines;
 };
 
 //! Base class for the \c Task<T> promise_type.
@@ -148,7 +149,7 @@ public:
      * This decides what should happen when the coroutine is finished.
      */
     auto final_suspend() const noexcept {
-        return TaskFinalSuspend{mAwaitingCoroutine};
+        return TaskFinalSuspend{mAwaitingCoroutines};
     }
 
     //! Called by co_await to obtain an Awaitable for type \c T.
@@ -225,13 +226,12 @@ public:
      *                          represented by this promise. When our coroutine finishes, it's
      *                          our job to resume the awaiting coroutine.
      */
-    bool setAwaitingCoroutine(std::coroutine_handle<> awaitingCoroutine) {
-        mAwaitingCoroutine = awaitingCoroutine;
-        return !mResumeAwaiter.exchange(true, std::memory_order_acq_rel);
+    void addAwaitingCoroutine(std::coroutine_handle<> awaitingCoroutine) {
+        mAwaitingCoroutines.push_back(awaitingCoroutine);
     }
 
     bool hasAwaitingCoroutine() const {
-        return mAwaitingCoroutine != nullptr;
+        return !mAwaitingCoroutines.empty();
     }
 
     bool setDestroyHandle() noexcept {
@@ -242,9 +242,7 @@ private:
     friend class TaskFinalSuspend;
 
     //! Handle of the coroutine that is currently co_awaiting this Awaitable
-    std::coroutine_handle<> mAwaitingCoroutine;
-    //! Indicates whether the awaiter should be resumed when it tries to co_await on us.
-    std::atomic<bool> mResumeAwaiter{false};
+    std::vector<std::coroutine_handle<>> mAwaitingCoroutines;
 
     //! Indicates whether we can destroy the coroutine handle
     std::atomic<bool> mDestroyHandle{false};
@@ -406,8 +404,8 @@ public:
      * co_awaited coroutine has finished synchronously and the co_awaiting coroutine doesn't
      * have to suspend.
      */
-    bool await_suspend(std::coroutine_handle<> awaitingCoroutine) noexcept {
-        return mAwaitedCoroutine.promise().setAwaitingCoroutine(awaitingCoroutine);
+    void await_suspend(std::coroutine_handle<> awaitingCoroutine) noexcept {
+        mAwaitedCoroutine.promise().addAwaitingCoroutine(awaitingCoroutine);
     }
 
 protected:
@@ -422,6 +420,20 @@ protected:
     std::coroutine_handle<Promise> mAwaitedCoroutine = {};
 };
 
+template<typename T>
+struct isTask : std::false_type {
+    using return_type = T;
+};
+
+template<typename T>
+struct isTask<QCoro::Task<T>> : std::true_type {
+    using return_type = typename QCoro::Task<T>::value_type;
+};
+
+template<typename T>
+constexpr bool isTask_v = isTask<T>::value;
+
+
 } // namespace detail
 
 /*! \endcond */
@@ -556,7 +568,8 @@ public:
     template<typename ThenCallback>
     requires (std::is_invocable_v<ThenCallback> || (!std::is_void_v<T> && std::is_invocable_v<ThenCallback, T>))
     auto then(ThenCallback &&callback) {
-        return thenImpl(std::forward<ThenCallback>(callback));
+        // Provide a custom error handler that simply re-throws the current exception
+        return thenImpl(std::forward<ThenCallback>(callback), [](const auto &) { throw; });
     }
 
     template<typename ThenCallback, typename ErrorCallback>
@@ -568,121 +581,63 @@ public:
 
 private:
     template<typename ThenCallback, typename ... Args>
-    requires (std::is_invocable_v<ThenCallback>)
-    auto invoke(ThenCallback &&callback, Args && ...) {
-        return callback();
+    auto invokeCb(ThenCallback &&callback, [[maybe_unused]] Args && ... args) {
+        if constexpr (std::is_invocable_v<ThenCallback, Args ...>) {
+            return callback(std::forward<Args>(args) ...);
+        } else {
+            return callback();
+        }
     }
 
     template<typename ThenCallback, typename Arg>
-    requires (std::is_invocable_v<ThenCallback, Arg>)
-    auto invoke(ThenCallback &&callback, Arg && arg) {
-        return callback(std::forward<Arg>(arg));
-    }
-
-    template<typename ThenCallback, typename Arg>
-    struct invoke_result: std::conditional_t<
+    struct cb_invoke_result: std::conditional_t<
         std::is_invocable_v<ThenCallback>,
             std::invoke_result<ThenCallback>,
             std::invoke_result<ThenCallback, Arg>
         > {};
 
     template<typename ThenCallback>
-    struct invoke_result<ThenCallback, void>: std::invoke_result<ThenCallback> {};
+    struct cb_invoke_result<ThenCallback, void>: std::invoke_result<ThenCallback> {};
 
     template<typename ThenCallback, typename Arg>
+    using cb_invoke_result_t = typename cb_invoke_result<ThenCallback, Arg>::type;
 
-    using invoke_result_t = typename invoke_result<ThenCallback, Arg>::type;
-
-    // Implementation of then() for callbacks that return Task<R>
-    template<typename ThenCallback, typename R = invoke_result_t<ThenCallback, T>>
-    requires QCoro::Awaitable<R>
-    auto thenImpl(ThenCallback &&callback) -> R {
-        const auto cb = std::move(callback);
-        if constexpr (std::is_void_v<value_type>) {
-            co_await *this;
-            co_return co_await invoke(cb);
-        } else {
-            co_return co_await invoke(cb, co_await *this);
+    template<typename R, typename ErrorCallback,
+             typename U = typename detail::isTask<R>::return_type>
+    auto handleException(ErrorCallback &errCb, const std::exception &exception) -> U {
+        errCb(exception);
+        if constexpr (!std::is_void_v<U>) {
+            return U{};
         }
     }
 
-    template<typename ThenCallback, typename ErrorCallback, typename R = invoke_result_t<ThenCallback, T>>
-    requires QCoro::Awaitable<R>
-    auto thenImpl(ThenCallback &&thenCallback, ErrorCallback &&errorCallback) -> R {
-        const auto thenCb = std::move(thenCallback);
-        const auto errCb = std::move(errorCallback);
+    template<typename ThenCallback, typename ErrorCallback, typename R = cb_invoke_result_t<ThenCallback, T>>
+    auto thenImpl(ThenCallback &&thenCallback, ErrorCallback &&errorCallback) -> std::conditional_t<detail::isTask_v<R>, R, Task<R>> {
+        auto thenCb = std::forward<ThenCallback>(thenCallback);
+        auto errCb = std::forward<ErrorCallback>(errorCallback);
         if constexpr (std::is_void_v<value_type>) {
             try {
                 co_await *this;
             } catch (const std::exception &e) {
-                errCb(e);
-                if constexpr (!std::is_void_v<typename R::value_type>) {
-                    co_return {};
-                } else {
-                    co_return;
-                }
+                co_return handleException<R>(errCb, e);
+            }
+            if constexpr (detail::isTask_v<R>) {
+                co_return co_await invokeCb(thenCb);
+            } else {
+                co_return invokeCb(thenCb);
             }
-            co_return co_await invoke(thenCb);
         } else {
-            std::optional<T> v;
+            std::optional<T> value;
             try {
-                v = co_await *this;
+                value = co_await *this;
             } catch (const std::exception &e) {
-                errCb(e);
-                if constexpr (!std::is_void_v<typename R::value_type>) {
-                    co_return {};
-                } else {
-                    co_return;
-                }
+                co_return handleException<R>(errCb, e);
             }
-            co_return co_await invoke(thenCb, std::move(*v));
-        }
-    }
-
-
-    // Implementation of then() for callbacks that return R, which is not Task<S>
-    template<typename ThenCallback, typename R = invoke_result_t<ThenCallback, T>>
-    requires (!QCoro::Awaitable<R>)
-    auto thenImpl(ThenCallback &&callback) -> Task<R> {
-        const auto cb = std::move(callback);
-        if constexpr (std::is_void_v<value_type>) {
-            co_await *this;
-            co_return invoke(cb);
-        } else {
-            co_return invoke(cb, co_await *this);
-        }
-    }
-
-    template<typename ThenCallback, typename ErrorCallback, typename R = invoke_result_t<ThenCallback, T>>
-    requires (!QCoro::Awaitable<R>)
-    auto thenImpl(ThenCallback &&thenCallback, ErrorCallback &&errorCallback) -> Task<R> {
-        const auto thenCb = std::move(thenCallback);
-        const auto errCb = std::move(errorCallback);
-        if constexpr (std::is_void_v<value_type>) {
-            try {
-                co_await *this;
-            } catch (const std::exception &e) {
-                errCb(e);
-                if constexpr (!std::is_void_v<R>) {
-                    co_return {};
-                } else {
-                    co_return;
-                }
+            if constexpr (detail::isTask_v<R>) {
+                co_return co_await invokeCb(thenCb, std::move(*value));
+            } else {
+                co_return invokeCb(thenCb, std::move(*value));
             }
-            co_return invoke(thenCb);
-        } else {
-            std::optional<T> v;
-            try {
-                v = co_await *this;
-            } catch (const std::exception &e) {
-                errCb(e);
-                if constexpr (!std::is_void_v<R>) {
-                    co_return {};
-                } else {
-                    co_return;
-                }
-            }
-            co_return invoke(thenCb, std::move(*v));
         }
     }
 
@@ -706,10 +661,42 @@ Task<void> inline TaskPromise<void>::get_return_object() noexcept {
     return Task<void>{std::coroutine_handle<TaskPromise>::from_promise(*this)};
 }
 
-} // namespace detail
 
 
-namespace detail {
+template <typename T>
+concept TaskConvertible = requires(T v, TaskPromiseBase t)
+{
+    { t.await_transform(v) };
+};
+
+template<typename T>
+struct awaitable_return_type {
+  using type = std::decay_t<decltype(std::declval<T>().await_resume())>;
+};
+
+template<QCoro::detail::has_member_operator_coawait T>
+struct awaitable_return_type<T> {
+    using type = std::decay_t<typename awaitable_return_type<decltype(std::declval<T>().operator co_await())>::type>;
+};
+
+template<QCoro::detail::has_nonmember_operator_coawait T>
+struct awaitable_return_type<T> {
+    using type = std::decay_t<typename awaitable_return_type<decltype(operator co_await(std::declval<T>()))>::type>;
+};
+
+template<Awaitable Awaitable>
+using awaitable_return_type_t = typename detail::awaitable_return_type<Awaitable>::type;
+
+template <typename Awaitable>
+requires TaskConvertible<Awaitable>
+using convertible_awaitable_return_type_t = typename detail::awaitable_return_type<decltype(std::declval<TaskPromiseBase>().await_transform(Awaitable()))>::type;
+
+template <typename Awaitable>
+requires TaskConvertible<Awaitable>
+auto toTask(Awaitable &&future) -> QCoro::Task<detail::convertible_awaitable_return_type_t<Awaitable>> {
+    co_return co_await future;
+}
+
 
 //! Helper class to run a coroutine in a nested event loop.
 /*!
@@ -720,32 +707,32 @@ namespace detail {
  * So instead we do basically what Qt does internally, but we make sure to not delete th
  * QFunctorSlotObjectWithNoArgs until after the event loop quits.
  */
-template<typename Coroutine>
-Task<> runCoroutine(QEventLoop &loop, bool &startLoop, Coroutine &&coroutine) {
-    co_await coroutine;
+template<Awaitable Awaitable>
+Task<> runCoroutine(QEventLoop &loop, bool &startLoop, Awaitable &&awaitable) {
+    co_await awaitable;
     startLoop = false;
     loop.quit();
 }
 
-template<typename T, typename Coroutine>
-Task<> runCoroutine(QEventLoop &loop, bool &startLoop, T &result, Coroutine &&coroutine) {
-    result = co_await coroutine;
+template<typename T, Awaitable Awaitable>
+Task<> runCoroutine(QEventLoop &loop, bool &startLoop, T &result, Awaitable &&awaitable) {
+    result = co_await awaitable;
     startLoop = false;
     loop.quit();
 }
 
-template<typename T, typename Coroutine>
-T waitFor(Coroutine &&coro) {
+template<typename T, Awaitable Awaitable>
+T waitFor(Awaitable &&awaitable) {
     QEventLoop loop;
     bool startLoop = true; // for early returns: calling quit() before exec() still starts the loop
     if constexpr (std::is_void_v<T>) {
-        runCoroutine(loop, startLoop, std::forward<Coroutine>(coro));
+        runCoroutine(loop, startLoop, std::forward<Awaitable>(awaitable));
         if (startLoop) {
             loop.exec();
         }
     } else {
         T result;
-        runCoroutine(loop, startLoop, result, std::forward<Coroutine>(coro));
+        runCoroutine(loop, startLoop, result, std::forward<Awaitable>(awaitable));
         if (startLoop) {
             loop.exec();
         }
@@ -766,13 +753,73 @@ T waitFor(Coroutine &&coro) {
  */
 template<typename T>
 inline T waitFor(QCoro::Task<T> &task) {
-    return detail::waitFor<T>(task);
+    return detail::waitFor<T>(std::forward<QCoro::Task<T>>(task));
 }
 
 // \overload
 template<typename T>
 inline T waitFor(QCoro::Task<T> &&task) {
-    return detail::waitFor<T>(task);
+    return detail::waitFor<T>(std::forward<QCoro::Task<T>>(task));
+}
+
+template<Awaitable Awaitable>
+inline auto waitFor(Awaitable &&awaitable) {
+    return detail::waitFor<detail::awaitable_return_type_t<Awaitable>>(std::forward<Awaitable>(awaitable));
+}
+
+//! Connect a callback to be called when the asynchronous task finishes.
+/*!
+ * Allows to register a callback to be called only when the context object
+ * still exists when the task finishes.
+ *
+ * In contrast to the then function, the result of connect can not be co_awaited.
+ *
+ * @param context A QObject that still needs to exist when the task finishes in order for the callback to be invoked.
+ * @param func The function that will be called when the task finishes.
+ *        For void tasks, it needs to have no arguments.
+ *        For all other types, it takes a value of the type as single argument.
+ */
+template <typename T, typename QObjectSubclass, typename Callback>
+requires std::is_invocable_v<Callback> || std::is_invocable_v<Callback, T> || std::is_invocable_v<Callback, QObjectSubclass *> || std::is_invocable_v<Callback, QObjectSubclass *, T>
+void connect(QCoro::Task<T> &&task, QObjectSubclass *context, Callback func) {
+    QPointer ctxWatcher = context;
+    if constexpr (std::is_same_v<T, void>) {
+        task.then([ctxWatcher, func = std::move(func)]() {
+            if (ctxWatcher) {
+                if constexpr (std::is_member_function_pointer_v<Callback>) {
+                    (ctxWatcher->*func)();
+                } else {
+                    func();
+                }
+            }
+        });
+    } else {
+        task.then([ctxWatcher, func = std::move(func)](auto &&value) {
+            if (ctxWatcher) {
+                if constexpr (std::is_invocable_v<Callback, QObjectSubclass, T>) {
+                    (ctxWatcher->*func)(std::forward<decltype(value)>(value));
+                } else if constexpr (std::is_invocable_v<Callback, T>) {
+                    func(std::forward<decltype(value)>(value));
+                } else {
+                    Q_UNUSED(value);
+                    if constexpr (std::is_member_function_pointer_v<Callback>) {
+                        (ctxWatcher->*func)();
+                    } else {
+                        func();
+                    }
+                }
+            }
+        });
+    }
+}
+
+template <typename T, typename QObjectSubclass, typename Callback>
+requires detail::TaskConvertible<T>
+        && (std::is_invocable_v<Callback> || std::is_invocable_v<Callback, detail::convertible_awaitable_return_type_t<T>> || std::is_invocable_v<Callback, QObjectSubclass *> || std::is_invocable_v<Callback, QObjectSubclass *, T>)
+        && (!detail::isTask_v<T>)
+void connect(T &&future, QObjectSubclass *context, Callback func) {
+    auto task = detail::toTask(std::move(future));
+    connect(std::move(task), context, func);
 }
 
 } // namespace QCoro
